#include <vector>
#include <fstream>
#include <string>
#include <vector>
#include <iostream>
#include <algorithm>
#include <map>
#include <cstdio>
#include <cerrno>

#include <sys/types.h>
#include <unistd.h>
#include <signal.h>

#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/trim.hpp>

#include <FL/Fl.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Check_Button.H>
#include <FL/Fl_Counter.H>
#include <FL/Fl_Group.H>
#include <FL/Fl_Return_Button.H>
#include <FL/Fl_Simple_Counter.H>
#include <FL/Fl_Text_Buffer.H>
#include <FL/Fl_Text_Display.H>
#include <FL/Fl_Window.H>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


const int win_width = 800;
const int win_height = 600;

const int margin = 4;

const int col_one = 110;
const int row_height = 32;

const int priorities_top = 20;

const int counter_width = 140;
const int counter_height = row_height - margin*2;

const int options_width = margin + col_one + counter_width + margin;

const int probe_width = 80;
const int col_two = 90;
const int inventory_width = col_two * 2 + probe_width * 2 + margin * 4;


Fl_Counter *storageWeight;
Fl_Counter *revenueWeight;
Fl_Counter *productionWeight;


Fl_Counter *iterations;
Fl_Counter *offsprings;
Fl_Counter *mutation;
Fl_Counter *population;
Fl_Counter *age;


std::vector<Fl_Simple_Counter*> inventory(20);

Fl_Window* sitesWindow = nullptr;


Fl_Text_Display* outputDisplay;
Fl_Text_Buffer* outputBuffer;

pid_t xenopid = 0;


const std::vector<std::string> probe_name = {
    "Mining G1",
    "Mining G2",
    "Mining G3",
    "Mining G4",
    "Mining G5",
    "Mining G6",
    "Mining G7",
    "Mining G8",
    "Mining G9",
    "Mining G10",
    "Research G1",
    "Research G2",
    "Research G3",
    "Research G4",
    "Research G5",
    "Research G6",
    "Booster G1",
    "Booster G2",
    "Storage",
    "Duplicator"
};


const std::vector<std::string> probe_code = {
    "M1", "M2", "M3", "M4", "M5", "M6", "M7", "M8", "M9", "M10",
    "R1", "R2", "R3", "R4", "R5", "R6", "B1", "B2", "S", "D"
};


struct Site {
    int max_sight;

    Fl_Check_Button* check = nullptr;
    Fl_Simple_Counter *counter = nullptr;

    Site(int ms) :
        max_sight{ms}
    {}
};

std::map<int, Site> sites = {
    {101,1},
    {102,0},
    {103,1},
    {104,1},
    {105,0},
    {106,1},
    {107,0},
    {108,0},
    {109,0},
    {110,1},
    {111,0},
    {112,0},
    {113,0},
    {114,0},
    {115,0},
    {116,0},
    {117,1},
    {118,0},
    {119,0},
    {120,0},
    {121,0},
    
    {201,0},
    {202,0},
    {203,0},
    {204,0},
    {205,0},
    {206,0},
    {207,0},
    {208,0},
    {209,0},
    {210,0},
    {211,0},
    {212,0},
    {213,1},
    {214,2},
    {215,0},
    {216,1},
    {217,0},
    {218,0},
    {219,0},
    {220,1},
    {221,2},
    {222,1},
    {223,1},
    {224,0},
    {225,1},
    
    {301,0},
    {302,0},
    {303,0},
    {304,0},
    {305,0},
    {306,1},
    {307,0},
    {308,0},
    {309,0},
    {310,0},
    {311,0},
    {312,0},
    {313,2},
    {314,0},
    {315,2},
    {316,0},
    {317,1},
    {318,2},
    {319,1},
    {320,0},
    {321,0},
    {322,0},
    
    {401,0},
    {402,0},
    {403,0},
    {404,1},
    {405,0},
    {406,0},
    {407,0},
    {408,1},
    {409,0},
    {410,1},
    {411,0},
    {412,0},
    {413,1},
    {414,2},
    {415,0},
    {416,0},
    {417,0},
    {418,0},
    {419,1},
    {420,0},
    
    {501,0},
    {502,1},
    {503,1},
    {504,0},
    {505,2},
    {506,1},
    {507,1},
    {508,1},
    {509,0},
    {510,0},
    {511,0},
    {512,0},
    {513,2},
    {514,1},
    {515,0},
    {516,0}
};



std::vector<std::vector<std::string>>
loadCSV(const std::string& fname)
{
    using std::vector;
    using std::string;

    vector<vector<string>> result;

    std::ifstream file{fname};
    if (!file)
        throw std::runtime_error{"could not open file " + fname};
    string line;
    while (getline(file, line)) {
        auto commentPos = line.find('#');
        if (commentPos != string::npos)
            line.erase(commentPos);
        boost::trim(line);
        if (line.empty())
            continue;
        vector<string> records;
        boost::split(records, line, [](char c) {return c==',';});
        result.push_back(std::move(records));
    }

    return result;
}


bool isSiteOp(const std::string& arg)
{
    if (arg.empty())
        return false;
    return arg[0] == '-' || arg[0] == '+';
}


void handleSite(const std::vector<std::string>& entry)
{
    if (entry[0].empty())
        return;
    char op;
    int name;
    
    if (std::sscanf(entry[0].c_str(), "%c%d", &op, &name) != 2) {
        std::cerr << "error parsing site update: " << entry[0] << std::endl;
        return;
    }
    switch (op) {
        case '-':
            sites.at(name).check->value(0);
            break;
        case '+':
            if (sites.at(name).counter)
                sites.at(name).counter->value(std::stoi(entry[1]));
            break;
        default:
            std::cerr << "invalid site operation: " << op << std::endl;
    }
}


void resetSites()
{
    for (auto& record : sites) {
        auto& site = record.second;
        site.check->value(1);
        if (site.counter)
            site.counter->value(site.max_sight);
    }
}



void loadInventory()
{
    
    try {
        auto data = loadCSV("inventory.txt");

        resetSites();
        
        for (auto& entry : data) {
            if (isSiteOp(entry[0])) {
                handleSite(entry);
            } else {
                if (data.size() < 2)
                    continue;
                auto code = entry[0];
                int num = std::stoi(entry[1]);
                if (num == 0)
                    continue;
                auto pos = std::find(probe_code.begin(),
                                     probe_code.end(),
                                     code);
                if (pos == probe_code.end()) {
                    std::cerr << "invalid probe code" << std::endl;
                    continue;
                }
                std::size_t idx = std::distance(probe_code.begin(), pos);

                inventory[idx]->value(num);
            }
        
        }
    }
    catch (...) {
        std::cerr << "error loading inventory" << std::endl;
    }
}



void saveInventory()
{
    try {
        std::ofstream output{"inventory.txt"};
        if (!output) {
            std::cerr << "error opening output file" << std::endl;
            return;
        }

        for (std::size_t idx=0; idx<inventory.size(); ++idx) {
            int val = static_cast<int>(inventory[idx]->value());
            if (val)
                output << probe_code[idx] << "," << val << std::endl;
        }

        for (auto& entry : sites) {
            int name = entry.first;
            const auto& site = entry.second;
            if (!site.check->value())
                output << "-" << name << std::endl;
            else if (site.counter && site.counter->value() < site.max_sight)
                output << "+" << name << "," << site.counter->value() << std::endl;
        }
    }
    catch (...) {
        std::cerr << "error writing inventory.txt" << std::endl;
    }
}


void closeSites(Fl_Widget* button)
{
    button->parent()->hide();
}



void buildSites()
{   
    int anchor_x = margin;
    int anchor_y = row_height;

    const int row_height = 26;
    
    const int check_width = 50;
    const int check_height = row_height - margin*2;
    const int counter_width = 60;
    const int group_width = 180;
    const int group_height = row_height * 25;

    const int window_width = (group_width + margin*2) * 5;
    const int window_height = 725;
    
    sitesWindow = new Fl_Window(window_width, window_height,
                                "Edit Sites");
    //sitesWindow->resizable(sitesWindow);


    auto populate = [&](int num)
    {
        std::string label = std::to_string(num);
        auto& site = sites.at(num);
        site.check = new Fl_Check_Button(anchor_x+margin, anchor_y+margin,
                                         check_width, check_height);
        site.check->copy_label(label.c_str());

        if (site.max_sight > 0) {
            site.counter = new Fl_Simple_Counter(anchor_x+margin + check_width + 60,
                                                 anchor_y+margin,
                                                 counter_width, check_height,
                                                 "Sight.:");
            site.counter->range(0, site.max_sight);
            site.counter->step(1);
            site.counter->align(FL_ALIGN_LEFT);
        }
    };

    Fl_Group* primordiaGroup = new Fl_Group(anchor_x, anchor_y,
                                            group_width, group_height,
                                            "Primordia");
    primordiaGroup->box(FL_DOWN_BOX);
    primordiaGroup->begin();
    for (int i=101; i<=121; ++i) {
        populate(i);
        anchor_y += row_height;
    }
    primordiaGroup->end();

    anchor_x += group_width + margin*2;
    anchor_y = row_height;
    Fl_Group* noctilumGroup = new Fl_Group(anchor_x, anchor_y,
                                           group_width, group_height,
                                           "Noctilum");
    noctilumGroup->box(FL_DOWN_BOX);
    noctilumGroup->begin();
    for (int i=201; i<=225; ++i) {
        populate(i);
        anchor_y += row_height;
    }
    noctilumGroup->end();

    anchor_x += group_width + margin*2;
    anchor_y = row_height;
    Fl_Group* obliviaGroup = new Fl_Group(anchor_x, anchor_y,
                                          group_width, group_height,
                                          "Oblivia");
    obliviaGroup->box(FL_DOWN_BOX);
    obliviaGroup->begin();
    for (int i=301; i<=322; ++i) {
        populate(i);
        anchor_y += row_height;
    }
    obliviaGroup->end();


    anchor_x += group_width + margin*2;
    anchor_y = row_height;
    Fl_Group* sylvalumGroup = new Fl_Group(anchor_x, anchor_y,
                                          group_width, group_height,
                                          "Sylvalum");
    sylvalumGroup->box(FL_DOWN_BOX);
    sylvalumGroup->begin();
    for (int i=401; i<=420; ++i) {
        populate(i);
        anchor_y += row_height;
    }
    sylvalumGroup->end();


    anchor_x += group_width + margin*2;
    anchor_y = row_height;
    Fl_Group* cauldrosGroup = new Fl_Group(anchor_x, anchor_y,
                                          group_width, group_height,
                                          "Cauldros");
    cauldrosGroup->box(FL_DOWN_BOX);
    cauldrosGroup->begin();
    for (int i=501; i<=516; ++i) {
        populate(i);
        anchor_y += row_height;
    }
    cauldrosGroup->end();
    
    Fl_Return_Button * okbutton = new Fl_Return_Button(window_width - 100 - margin*2,
                                                       window_height - 30 - margin*2,
                                                       100, 30,
                                                       "OK");
    okbutton->callback(closeSites);
    
    sitesWindow->end();



    resetSites();
}





void showSites()
{
    sitesWindow->set_modal();
    sitesWindow->show();
}


void addLine(const std::string& line)
{
    static bool firstTime = true;
    if (firstTime) {
        outputBuffer->append(line.c_str());
        firstTime = false;
    } else {
        std::string witheol = "\n" + line;
        outputBuffer->append(witheol.c_str());
    }
    outputDisplay->scroll(INT_MAX, 0);
}


std::string buildCommand()
{
    std::string cmd;

#ifndef __MINGW32__
    cmd += "./xenoprobes";
#else
    cmd += "xenoprobes.exe";
#endif

    cmd += " --storageweight " + std::to_string(int(storageWeight->value()));
    cmd += " --revenueweight " + std::to_string(int(revenueWeight->value()));
    cmd += " --productionweight " + std::to_string(int(productionWeight->value()));

    cmd += " --iterations " + std::to_string(int(iterations->value()));
    cmd += " --offsprings " + std::to_string(int(offsprings->value()));
    cmd += " --population " + std::to_string(int(population->value()));
    cmd += " --age " + std::to_string(int(age->value()));
    cmd += " --pipe";

    return cmd;
}



void runOptimizer()
{
    saveInventory();
    
    std::string cmdline = buildCommand();

    addLine("Running:\n\t" + cmdline);

    FILE *xenopipe = popen(cmdline.c_str(), "r");
    if (!xenopipe) {
        int errorcode = errno;
        std::string msg = std::strerror(errorcode);
        addLine("Error launching process: " + msg);
        return;
    }

    static char linebuffer[1024];

    if (!std::fgets(linebuffer, sizeof linebuffer, xenopipe)) {
        addLine("Error: could not read subprocess PID");
        return;
    }
    try {
        std::string line = linebuffer;
        xenopid = std::stoll(line);
    }
    catch (...) {
        addLine("Error: could not convert the PID");
        return;
    }


    while (std::fgets(linebuffer, sizeof linebuffer, xenopipe)) {
        std::string line = linebuffer;
        boost::trim(line);
        addLine(line);
        Fl::wait(0.05);
    }

    pclose(xenopipe);
}


void stopOptimizer()
{
    #ifndef __MINGW32__
    if (xenopid) {
        kill(xenopid, SIGINT);
        xenopid = 0;
        addLine("Stopped optimizer");
    }
    #endif
}





int main(int argc, char **argv)
{
    const char* title = "XenoProbes GUI"
    #ifdef PACKAGE_STRING
    " (" PACKAGE_STRING ")"
    #endif
                        ;

    Fl_Window* window = new Fl_Window(win_width, win_height, title);

    window->resizable(window);


    // PRIORITIES
    int anchor_x = margin;
    int anchor_y = row_height;
    
    Fl_Group* priorityGroup = new Fl_Group(anchor_x,
                                           anchor_y,
                                           options_width,
                                           row_height * 3,
                                           "Priorities");
    priorityGroup->align(FL_ALIGN_TOP_LEFT);
    priorityGroup->box(FL_DOWN_BOX);
    priorityGroup->begin();
    
    anchor_x += col_one;
    anchor_y += margin;
    
    storageWeight = new Fl_Counter(anchor_x, anchor_y,
                                   counter_width, counter_height,
                                   "Storage");
    storageWeight->align(FL_ALIGN_LEFT);
    storageWeight->step(1, 100);
    storageWeight->value(1000);


    anchor_y += row_height;
    
    revenueWeight = new Fl_Counter(anchor_x, anchor_y,
                                   counter_width, counter_height,
                                   "Revenue");
    revenueWeight->align(FL_ALIGN_LEFT);
    revenueWeight->step(1, 100);
    revenueWeight->value(10);

    
    anchor_y += row_height;

    productionWeight = new Fl_Counter(anchor_x, anchor_y,
                                      counter_width, counter_height,
                                      "Production");
    productionWeight->align(FL_ALIGN_LEFT);
    productionWeight->step(1, 100);
    productionWeight->value(1);

    priorityGroup->end();





    // OPTIONS
    anchor_x = margin;
    anchor_y += 4*margin + row_height;

    Fl_Group *optionsGroup = new Fl_Group(anchor_x, anchor_y,
                                          options_width,
                                          row_height * 5,
                                          "Options");
    optionsGroup->align(FL_ALIGN_TOP_LEFT);
    optionsGroup->box(FL_DOWN_BOX);
    optionsGroup->begin();
    
    anchor_x += col_one;
    anchor_y += margin;

    iterations = new Fl_Counter(anchor_x, anchor_y,
                                counter_width, counter_height,
                                "Iterations");
    iterations->align(FL_ALIGN_LEFT);
    iterations->step(1, 100);
    iterations->value(2000);
    iterations->minimum(1);    

    
    anchor_y += row_height;
    
    offsprings = new Fl_Counter(anchor_x, anchor_y,
                                counter_width, counter_height,
                                "Offsprings");
    offsprings->align(FL_ALIGN_LEFT);
    offsprings->step(1, 100);
    offsprings->value(100);
    offsprings->minimum(1);



    anchor_y += row_height;
    
    mutation = new Fl_Counter(anchor_x, anchor_y,
                              counter_width, counter_height,
                              "Mutation rate");
    mutation->align(FL_ALIGN_LEFT);
    mutation->step(0.001, 0.1);
    mutation->value(0.03125);
    mutation->range(0, 1);



    anchor_y += row_height;
    
    population = new Fl_Counter(anchor_x, anchor_y,
                                counter_width, counter_height,
                                "Population");
    population->align(FL_ALIGN_LEFT);
    population->step(1, 100);
    population->value(200);
    population->minimum(1);


    anchor_y += row_height;
    
    age = new Fl_Counter(anchor_x, anchor_y,
                         counter_width, counter_height,
                         "Age");
    age->align(FL_ALIGN_LEFT);
    age->step(1, 100);
    age->value(50);
    age->minimum(0);

    
    optionsGroup->end();




    anchor_x = margin + options_width + margin + margin;
    anchor_y = row_height;
    

    // INVENTORY
    
    
    Fl_Group* inventoryGroup = new Fl_Group(anchor_x, anchor_y,
                                            inventory_width,
                                            row_height * 11,
                                            "Inventory");
    inventoryGroup->align(FL_ALIGN_TOP_LEFT);
    inventoryGroup->box(FL_DOWN_BOX);
    inventoryGroup->begin();
    
    anchor_x += margin + col_two;
    anchor_y += margin;

    for (int i=0; i<10; ++i) {
        inventory[i] = new Fl_Simple_Counter(anchor_x, anchor_y,
                                             probe_width, counter_height,
                                             probe_name[i].c_str());
        inventory[i]->align(FL_ALIGN_LEFT);
        inventory[i]->step(1);
        inventory[i]->value(0);
        inventory[i]->minimum(0);

        anchor_y += row_height;
    }


    anchor_y = row_height + margin;
    anchor_x += probe_width + margin * 2 + col_two;

    for (int i=10; i<20; ++i) {
        inventory[i] = new Fl_Simple_Counter(anchor_x, anchor_y,
                                             probe_width, counter_height,
                                             probe_name[i].c_str());
        inventory[i]->align(FL_ALIGN_LEFT);
        inventory[i]->step(1);
        inventory[i]->value(0);
        inventory[i]->minimum(0);

        anchor_y += row_height;
    }


    anchor_x = margin + options_width + margin + margin + margin;
    
    Fl_Button* loadInv = new Fl_Button(anchor_x, anchor_y,
                                       probe_width, counter_height,
                                       "Load");
    loadInv->callback(reinterpret_cast<Fl_Callback*>(loadInventory));



    Fl_Button* editSites = new Fl_Button(anchor_x + options_width/2 - probe_width/2,
                                         anchor_y,
                                         probe_width, counter_height,
                                         "Edit Sites");
    editSites->callback(reinterpret_cast<Fl_Callback*>(showSites));

    
    anchor_x += inventory_width - margin - margin - probe_width;
    
    Fl_Button* saveInv = new Fl_Button(anchor_x, anchor_y,
                                       probe_width, counter_height,
                                       "Save");
    saveInv->callback(reinterpret_cast<Fl_Callback*>(saveInventory));
    
    inventoryGroup->end();


    anchor_y = row_height * 12 + margin * 4;

    Fl_Button* optimize = new Fl_Button(margin * 2,
                                        anchor_y,
                                        win_width - margin*4 - 80 - margin*2,
                                        40,
                                        "Save inventory and OPTIMIZE!");
    optimize->callback(reinterpret_cast<Fl_Callback*>(runOptimizer));
    optimize->labelsize(24);

#ifndef __MINGW32__
    Fl_Button* stop = new Fl_Button(margin * 2 + win_width - margin*4 - 80 - margin,
                                    anchor_y,
                                    80,
                                    40,
                                    "STOP!");
    stop->callback(reinterpret_cast<Fl_Callback*>(stopOptimizer));
    stop->labelsize(24);
#else
    optimize->resize(margin * 2, anchor_y,
                     win_width - margin*4,
                     40);
#endif


    
    anchor_y += 46;
    
    
    outputBuffer = new Fl_Text_Buffer;
    outputDisplay = new Fl_Text_Display(margin,
                                        anchor_y,
                                        win_width - margin*2,
                                        win_height - margin - anchor_y);
    outputDisplay->buffer(outputBuffer);
    outputDisplay->wrap_mode(Fl_Text_Display::WRAP_AT_BOUNDS, 0);
    
    window->resizable(outputDisplay);
    
    
    window->end();


    buildSites();
    
    
    window->show(argc, argv);
    
    loadInventory();

    return Fl::run();
}
